unit rDBComponents_UDB;

{<
  rDBComponents_UDB (UDB= Universal DataBase) is set of enhanced DB components
  which can be used instead of standard db components and provide
  the same properties and functions for different types of databases.
  It allows to convert application very easilly from one database type to another.

  Now it supports following type of database connections and queries:
  ADO = dbGo (used for MSAccess, MSSQLServer nad other db via ODBC or JET)
  ZEOS (used for MySQL, Firebird and other supported databases)
  DBX = dbExpress (used for MySQL, Interbase, Firebird and other supported databases)
  IB = Interbase (used for Interbase or Firebird)
  FIB = FIBBlus  (used for Interbase or Firebird)
  FireDac

  Usage of particular DB type is set by compiler conditional defines:
  DEFINE USEADO  for ADO
  DEFINE USEZEOS for Zeos
  DEFINE USEDBX  for DBX
  DEFINE USEIB   for Interbase
  DEFINE USEFIB  for FIBBlus
  DEFINE USEFD  for FireDac


  Unit contains new data connection components:
  TConnectionUDB - descendant of TADOConnection/TZConnection/TSQLConnection/TIBDataBase/TpFIBDatabase/TFDConnection
  TQueryUDB - descendant of TADOQuery/TZQuery/TSimpleDataSet/TIBQuery/TpFIBDataSet/TFDQuery

  TQueryUDB provides many new functions and properties usefull for any
  db application for centralization of all db tasks:
  - new props (TableName, IdFieldName, FieldTagToDuplic, FieldTagToReadOnly...)
  - new function (DuplicateRecord, SetFieldsReadOnly - according of fields tag value)
  - new events for centralized setting of fields captions and properties in application
  - many new function for setting of SQL commands in runtime


  Unit contain set of visible DB components.
  All of them contain internal TQueryUDB dataset with definable SQL command
  for loading of list values and events for setting of SQL in run time,
  so no other auxilary dataset is needed on the form.
  It can show bounded buttons used for enhanced selection of value
  or showing details of selected record dependently on application.

  TEditUDB - TEdit descendant loads and display text from any SQL command
  TComboBoxUDB - TrComboBoxEx descendant load list of items from any SQL command
  TrDBComboBoxUDB - TrDBComboBox descendant load list of items from any SQL command
  TrDBLookupComboBoxUDB - TrDBLookupComboBox descendant load list of items from any SQL command
  TListBoxUDB - TListBox descendant load list of items from any SQL command
  TrDBListBoxUDB - TrDBListBox descendant load list of items from any SQL command
  TrDBLookupListBoxUDB - TrDBLookupListBox descendant load list of items from any SQL command

  TCheckListBoxUDB - TCheckListBox descendant with possibility to load list of items from any SQL command
  TrDBCheckListBoxUDB - TrDBCheckListBox descendant with possibility to load list of items and values
  from any SQL command, provide all functions for N:M relationship

  TrDBRecViewUDB - TrDBRecView descendant with possibility to load/show/edit record from any SQL command
  TrDBGrid_MSUDB - TrDBGrid_MS descendant with possibility to load/show/edit records from any SQL command

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2011)
  @lastmod(Apr 2017)
}

interface

{$DEFINE DoNotUseGlyphInDesign}
{$IF Defined(USEADO) OR Defined(USEDBX) OR Defined(USEIB) OR Defined(USEFIB) OR Defined(USEFD)}
{$DEFINE AddConReadOnly}
{$DEFINE AddConHostName}
{$IFEND}
{$IF Defined(USEADO) OR Defined(USEIB) OR Defined(USEFIB) OR Defined(USEFD)}
{$DEFINE AddQueryReadOnly}
{$IFEND}
{$IF Defined(USEDBX) OR Defined(USEFIB)}
{$DEFINE AddQuerySQL}
{$IFEND}
{$IF Defined(USEIB) OR Defined(USEFIB)}
{$DEFINE AddQueryConnection}
{$IFEND}
uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus, Dialogs,
  StdCtrls, ExtCtrls, Buttons, Forms, DB, Variants, CheckLst,
{$IFDEF USEADO} Adodb, {$ENDIF}
{$IFDEF USEZeos} ZConnection, ZAbstractRODataset, ZAbstractDataset, ZDataSet, {$ENDIF}
{$IFDEF USEDBX} SqlExpr, DBClient, SimpleDS, {$ENDIF}
{$IFDEF USEIB} IBDatabase, IBCustomDataSet, IBQuery, {$ENDIF}
{$IFDEF USEFIB} FIBDatabase, pFIBDatabase, FIBDataSet, pFIBDataSet, {$ENDIF}
{$IFDEF USEFD} FireDAC.Comp.DataSet, FireDAC.Comp.Client, {$ENDIF}
  rDBComponents, rImprovedComps, rDBRecView, rDBGrid, rDBGrid_MS;

type
  TConnectionUDB = class;
  TQueryUDB = class;

  { list of dataset event types }
  TDatasetEventType = (deNone, deBeforeOpen, deAfterOpen, deBeforeClose, deAfterClose, deBeforePost, deAfterPost, deBeforeScroll, deAfterScroll,
    deBeforeCancel, deAfterCancel, deBeforeInsert, deAfterInsert, deBeforeEdit, deAfterEdit, deBeforeDelete, deAfterDelete, deBeforeApplyUpdates,
    deAfterApplyUpdates, deAfterPostOrApplyUpdates);
  { set of dataset event types}
  TDatasetEventTypes = set of TDatasetEventType;

  { event type used by UDBComponents }
  TDataSetEvent = procedure(DataSet: TDataSet; EventType: TDatasetEventType = deNone) of object;
  { event type used by TQueryUDB to get SQL command }
  TGetSQLEvent = procedure(Sender: TObject; SQL: TStrings) of object;
  { event type used by UDBComponents after click on bound button }
  TClickComboButton = procedure(Sender: TObject; Con: TConnectionUDB; TableName: string; DataField, ListField: TField) of object;

  { TConnectionUDB - descendant of TADOConnection/TZConnection/TSQLConnection/TIBDataBase/TpFIBDatabase }
  TConnectionUDB = class(
{$IFDEF USEADO} TADOConnection {$ENDIF}
{$IFDEF USEZeos} TZConnection {$ENDIF}
{$IFDEF USEDBX} TSQLConnection {$ENDIF}
{$IFDEF USEIB} TIBDataBase {$ENDIF}
{$IFDEF USEFIB} TpFIBDatabase {$ENDIF}
{$IFDEF USEFD} TFDConnection {$ENDIF})
  private
{$IFDEF USEADO}
    FConnectionStringOrig: string;
    FConnectionStringDesign: string;
{$ENDIF}
{$IFDEF AddConReadOnly}
    FReadOnly: boolean;
    procedure SetReadOnly(Value: boolean);
{$ENDIF}
{$IFDEF AddConHostName}
    function GetHostName: string;
{$ENDIF}
  protected
{$IFDEF USEADO}
    procedure DoConnect; override;
{$ENDIF}
  public
    { return database name }
    function GetDatabaseName: string;
    { scan all registered dataset and call CallBackProc for each one }
    procedure CallProcForRegisteredDataset(CallBackProc: TNotifyEvent);
{$IFDEF USEZeos}
    { open connection }
    procedure Open;
    { close connection }
    procedure Close;
{$ENDIF}
  published
{$IFDEF USEADO}
    { copy of connection string in design time }
    property ConnectionStringDesign: string read FConnectionStringDesign write FConnectionStringDesign;
{$ENDIF}
{$IFDEF AddConReadOnly}
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
{$ENDIF}
{$IFDEF AddConHostName}
    property HostName: string read GetHostName stored false;
{$ENDIF}
{$IFDEF USEZeos}
    { list of registered datasets }
    property RegisteredDatasets: TList read FDataSets;
{$ENDIF}
  end;

  { TQueryUDB - descendant of TADOQuery/TZQuery/TSimpleDataSet/TIBQuery/TpFIBDataSet }
  TQueryUDB = class(
{$IFDEF USEADO} TADOQuery {$ENDIF}
{$IFDEF USEZeos} TZQuery {$ENDIF}
{$IFDEF USEDBX} TSimpleDataSet {$ENDIF}
{$IFDEF USEIB} TIBQuery {$ENDIF}
{$IFDEF USEFIB} TpFIBDataSet {$ENDIF}
{$IFDEF USEFD} TFDQuery {$ENDIF})
  private
    FTableName: string;
    FIdFieldName: string;
    FFieldImgIdx: TStrings;
    FFieldPropsAutoLoad: boolean;
    FFieldDefFormatAutoSet: boolean;
    FFieldTagToDuplic: byte;
    FFieldTagToReadOnly: byte;
    FConfirmDelete: boolean;
    FCheckFieldSizeBeforePost: boolean;
    FLastSort: string;
    FDataSetEvent: TDataSetEvent;
    FEventNotifyArr: array of TDataSetEvent;
    FControlsToUpdate: TStrings;
    FControlsReadOnlyColor: TColor;
    FControlsUpdating: boolean;
{$IFDEF AddQueryReadOnly}
    FReadOnly: boolean;
{$ENDIF}
{$IFDEF AddQuerySQL}
    FSQL: TStrings;
{$ENDIF}
{$IFDEF USEDBX}
    FAutoApplyUpdates: boolean;
    procedure IntApplyUpdates;
    procedure IntReconcileError(DataSet: TCustomClientDataSet; E: EReconcileError; UpdateKind: TUpdateKind; var Action: TReconcileAction);
{$ENDIF}
    procedure SetFieldImgIdx(Value: TStrings);
    procedure SetControlsToUpdate(Value: TStrings);
{$IFDEF AddQueryReadOnly}
    function GetReadOnly: boolean;
{$ENDIF}
{$IFDEF AddQuerySQL}
    function GetSQL: TStrings;
    procedure SetSQL(Value: TStrings);
    procedure SQLChanged(Sender: TObject);
{$ENDIF}
{$IFDEF AddQueryConnection}
    function GetConnection: {$IFDEF USEIB} TIBDataBase {$ENDIF} {$IFDEF USEFIB} TFIBDatabase {$ENDIF};
    procedure SetConnection(Value: {$IFDEF USEIB} TIBDataBase {$ENDIF} {$IFDEF USEFIB} TFIBDatabase {$ENDIF});
{$ENDIF}
    procedure AddIntLog(const act: string; AddRecCount: boolean);
    procedure ControlsToUpdateChange(Sender: TObject);
  protected
    procedure Loaded; override;
{$IFDEF AddQueryReadOnly}
    function GetCanModify: boolean; override;
{$ENDIF}
    procedure DoBeforeOpen; override;
    procedure DoAfterOpen; override;
    procedure DoBeforeClose; override;
    procedure DoAfterClose; override;
    procedure DoBeforePost; override;
    procedure DoAfterPost; override;
    procedure DoBeforeScroll; override;
    procedure DoAfterScroll; override;
    procedure DoBeforeCancel; override;
    procedure DoAfterCancel; override;
    procedure DoBeforeInsert; override;
    procedure DoAfterInsert; override;
    procedure DoBeforeEdit; override;
    procedure DoAfterEdit; override;
    procedure DoBeforeDelete; override;
    procedure DoAfterDelete; override;
{$IFDEF USEDBX}
    procedure DoAfterApplyUpdates(var OwnerData: OleVariant); override;
    procedure DoBeforeApplyUpdates(var OwnerData: OleVariant); override;
{$ENDIF}
    procedure CallDatasetEvent(Event: TDatasetEventType);
    procedure ProcControlsToUpdate(Event: TDatasetEventType);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Post; override; {< @exclude }
    procedure SetReadOnly(Value: boolean);
    procedure SetSortEx(const ASort: string);
    procedure RequeryEx(ChangeCursor: boolean = true);
    procedure SetSQLWhere(const Condition: string); overload;
    procedure SetSQLWhere(const FormatCondition: string; Value: variant); overload;
    procedure SetSQLWhere(Field: TField); overload;
    procedure SetSQLWhere(Field: TField; Value: variant); overload;
    procedure AddSQLWhereOR(const Condition: string);
    procedure AddSQLWhereAND(const Condition: string);
    function SetExecSQL(const SQLCmd: string): integer;
    procedure SetFieldDefaultFormat;
    procedure LoadFieldsProps;
{$IFDEF USEZeos}
    property ResultSet;
    function FindMetadataColumnIndex(const FieldName: string): integer;
{$ENDIF}
    function GetTableName: string;
    procedure UpdateDatasetState;
    function IsEditing: boolean;
    function IsNewRec: boolean;
    function HasFieldToDuplic: boolean;
    function SetFieldsReadOnly: integer;
    function DuplicateRecord: boolean;
    procedure AddEventNotify(CallBackProc: TDataSetEvent);
    procedure RemoveEventNotify(CallBackProc: TDataSetEvent);
  published
    property DataSetField;
    property TableName: string read FTableName write FTableName;
    property IdFieldName: string read FIdFieldName write FIdFieldName;
    property FieldImgIdx: TStrings read FFieldImgIdx write SetFieldImgIdx;
    property FieldPropsAutoLoad: boolean read FFieldPropsAutoLoad write FFieldPropsAutoLoad default true;
    property FieldDefFormatAutoSet: boolean read FFieldDefFormatAutoSet write FFieldDefFormatAutoSet default true;
    property ControlsToUpdate: TStrings read FControlsToUpdate write SetControlsToUpdate;
    property ControlsReadOnlyColor: TColor read FControlsReadOnlyColor write FControlsReadOnlyColor default clBtnFace;
    property ConfirmDelete: boolean read FConfirmDelete write FConfirmDelete default false;
{$IFDEF AddQueryReadOnly}
    property ReadOnly: boolean read GetReadOnly write SetReadOnly default false;
{$ENDIF}
    property FieldTagToDuplic: byte read FFieldTagToDuplic write FFieldTagToDuplic default 1;
    property FieldTagToReadOnly: byte read FFieldTagToReadOnly write FFieldTagToReadOnly default 2;
    property CheckFieldSizeBeforePost: boolean read FCheckFieldSizeBeforePost write FCheckFieldSizeBeforePost default false;
{$IFDEF AddQuerySQL}
    property SQL: TStrings read GetSQL write SetSQL;
{$ENDIF}
{$IFDEF USEDBX}
    property AutoApplyUpdates: boolean read FAutoApplyUpdates write FAutoApplyUpdates default true;
{$ENDIF}
{$IFDEF AddQueryConnection}
    property Connection: {$IFDEF USEIB} TIBDataBase {$ENDIF} {$IFDEF USEFIB} TFIBDatabase {$ENDIF} read GetConnection write SetConnection;
{$ENDIF}
    property OnDataSetEvent: TDataSetEvent read FDataSetEvent write FDataSetEvent;
  end;

  { TQueryUDB descendand for internal purpose only }
  { @exclude }
  TQueryIntList = class(TQueryUDB)
  published
    constructor Create(AOwner: TComponent); override; {< @exclude }
  end;

  { TSpeedButton descendand for internal purpose only as bounded button }
  { @exclude }
  TBoundComboButton = class(TSpeedButton)
  private
    ParentW: TControl;
    FIdx: integer;
    GlyphResourceName: string;
    procedure SetIdx(Value: integer);
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean{$ENDIF}); override;
    procedure Loaded; override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
  public
    constructor CreateBtn(AOwner: TComponent; AName, ACaption, AGlyphResourceName: string; Idx: integer); virtual;
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    procedure RefreshPosition;
    procedure Click; override;
    property Idx: integer read FIdx write SetIdx;
  published
    property Left stored false;
    property Top stored false;
    property Width stored false;
    property Height stored false;
  end;

  { TrDBCheckBox descendand with boolean value 1 and 0 }
  { @exclude }
  TrDBCheckBoxUDB = class(TrDBCheckBox)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  { TEdit descendant loads and display text from any SQL command }
  TEditUDB = class(TEdit)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh text value }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
  published
    { define SQL command, value of first field is used as text value }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TComboBox descendant load list of items from any SQL command }
  TComboBoxUDB = class(TrComboBoxEx)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FDelKeyForNullValue: boolean;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected item (record) }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define whether DEL key set null value }
    property DelKeyForNullValue: boolean read FDelKeyForNullValue write FDelKeyForNullValue default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TrDBEdit descendant show bound button }
  TrDBEditUDB = class(TrDBEdit)
  private
    FEditButton: TBoundComboButton;
    FEditButtonVisible: boolean;
    FEditButtonAllwaysEnabled: boolean;
    procedure SetEditButtonVisible(Value: boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
  published
    { define aux button beside control }
    property EditButton: TBoundComboButton read FEditButton;
    { define visibility of aux button beside control }
    property EditButtonVisible: boolean read FEditButtonVisible write SetEditButtonVisible default true;
    { define enability of aux button beside control }
    property EditButtonAllwaysEnabled: boolean read FEditButtonAllwaysEnabled write FEditButtonAllwaysEnabled default true;
  end;

  { TrDBComboBox descendant load list of items from any SQL command }
  TrDBComboBoxUDB = class(TrDBComboBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TConnectionUDB;
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TrDBLookupComboBox descendant load list of items from any SQL command }
  TrDBLookupComboBoxUDB = class(TrDBLookupComboBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDetailButton: TBoundComboButton;
    FListDetailButtonVisible: boolean;
    FClickComboDetailButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TConnectionUDB;
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDetailButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    function GetListValue: string;
    function GetListDetailButtonEnability: boolean;
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
    procedure KeyValueChanged; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { indicate whether list dataset as active }
    function ListActive: boolean;
  published
    { define SQL command, first field is used as keyvalue field,
      2nd field as list value field }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;

    { define 2nd aux button beside control }
    property ListDetailButton: TBoundComboButton read FListDetailButton;
    { define visibility of 2nd aux button beside control }
    property ListDetailButtonVisible: boolean read FListDetailButtonVisible write SetListDetailButtonVisible default true;
    { define event after ListDetailButtonClick }
    property OnClickComboDetailButton: TClickComboButton read FClickComboDetailButton write FClickComboDetailButton;
    { return list value for selected item }
    property ListValue: string read GetListValue;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TListBox descendant load list of items from any SQL command }
  TListBoxUDB = class(TrListBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TrDBListBox descendant load list of items from any SQL command }
  TrDBListBoxUDB = class(TrDBListBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TConnectionUDB;
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TrDBLookupListBox descendant load list of items from any SQL command }
  TrDBLookupListBoxUDB = class(TrDBLookupListBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TConnectionUDB;
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    function GetListValue: string;
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage);
    message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, first field is used as keyvalue field,
      2nd field as list value field }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { return list value for selected item }
    property ListValue: string read GetListValue;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TCheckListBox descendant with possibility to load list of items from any SQL command }
  TCheckListBoxUDB = class(TrCheckListBox)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as text value }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
  end;

  { TrDBCheckListBox descendant with possibility to load list of items and values
    from any SQL command, provide all functions for N:M relationship }
  TrDBCheckListBoxUDB = class(TrDBCheckListBox)
  private
    FListConnection: TConnectionUDB;
    FItemsListDataSet: TQueryIntList;
    FItemsListSQL: TStrings;
    FItemsListAutoLoad: boolean;
    FItemsListSQLTryNow: boolean;
    FGetItemsListSQL: TGetSQLEvent;
    FItemsValueDataSet: TQueryIntList;
    FItemsValueSQL: TStrings;
    FItemsValueSQLTryNow: boolean;
    FGetItemsValueSQL: TGetSQLEvent;
    FItemsValueDataSetEvent: TDataSetEvent;
    FLinkedDSEvent: TList;
    FOnBeforeSaveItemValues: TLoadSaveItemList;
    function GetListConnection: TConnectionUDB;
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetItemsListSQL(Value: TStrings);
    procedure SetItemsListSQLTryNow(Value: boolean);
    procedure SetItemsValueSQL(Value: TStrings);
    procedure SetItemsValueSQLTryNow(Value: boolean);
    procedure SetItemsValueDataSetEvent(Value: TDataSetEvent);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure DoLoadItemValues; override; {< @exclude }
    procedure DoSaveItemValues; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItemsList(ReloadValues: boolean);
    { open SQL command and refresh list values = checkboxes }
    procedure ReloadItemsValue;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define SQL command for loading of list items,
      value of first field is used as N-table ID value, value of 2nd field as N-table listed text }
    property ItemsListSQL: TStrings read FItemsListSQL write SetItemsListSQL;
    { define whether ReloadItemsList is called automatically after loading }
    property ItemsListAutoLoad: boolean read FItemsListAutoLoad write FItemsListAutoLoad default true;
    { try to open ItemsListSQL command in design time }
    property ItemsListSQLTryNow: boolean read FItemsListSQLTryNow write SetItemsListSQLTryNow default false;
    { call before opening of ItemsListSQL to allow update command or set WHERE condition }
    property OnGetItemsListSQL: TGetSQLEvent read FGetItemsListSQL write FGetItemsListSQL;
    { define SQL command for loading of item values,
      value of first field is used as M-table ID value, value of 2nd field as N-table ID value }
    property ItemsValueSQL: TStrings read FItemsValueSQL write SetItemsValueSQL;
    { try to open ItemsValueSQL command in design time }
    property ItemsValueSQLTryNow: boolean read FItemsValueSQLTryNow write SetItemsValueSQLTryNow default false;
    { call before opening of ItemsValueSQL to allow update command or set WHERE condition }
    property OnGetItemsValueSQL: TGetSQLEvent read FGetItemsValueSQL write FGetItemsValueSQL;
    { return internal List Dataset to allow reading of other fields }
    property ItemsListDataSet: TQueryIntList read FItemsListDataSet;
    { return internal Value Dataset to allow reading of other fields }
    property ItemsValueDataSet: TQueryIntList read FItemsValueDataSet;
    { define event called before saving of items }
    property OnBeforeSaveItemValues: TLoadSaveItemList read FOnBeforeSaveItemValues write FOnBeforeSaveItemValues;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnItemsValueDataSetEvent: TDataSetEvent read FItemsValueDataSetEvent write SetItemsValueDataSetEvent;
  end;

  { TrDBRecView descendant with possibility to load/show/edit record from any SQL command }
  TrDBRecViewUDB = class(TrDBRecView)
  private
    FRecSQL: TStrings;
    FRecConnection: TConnectionUDB;
    FRecDataSet: TQueryIntList;
    FRecDataSource: TDataSource;
    FTemplateDataSource: TDataSource;
    FRecDataSetReadOnly: boolean;
    FRecAutoLoad: boolean;
    FRecSQLTryNow: boolean;
    FOnGetRecSQL: TGetSQLEvent;
    FRecDataSetEvent: TDataSetEvent;
    procedure SetRecSQL(Value: TStrings);
    procedure SetRecConnection(Value: TConnectionUDB);
    procedure SetRecSQLTryNow(Value: boolean);
    procedure SetRecDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh recview content }
    procedure ReloadData;
    { return internal Dataset to allow reading of other fields }
    property RecDataSet: TQueryIntList read FRecDataSet;
    { return internal Datasource to allow using for other DB components }
    property RecDataSource: TDataSource read FRecDataSource;
  published
    { define SQL command }
    property RecSQL: TStrings read FRecSQL write SetRecSQL;
    { define UDB connection }
    property RecConnection: TConnectionUDB read FRecConnection write SetRecConnection;
    { define whether ReloadData is called automatically after loading }
    property RecAutoLoad: boolean read FRecAutoLoad write FRecAutoLoad default true;
    { try to open SQL command in design time }
    property RecSQLTryNow: boolean read FRecSQLTryNow write SetRecSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetRecSQL: TGetSQLEvent read FOnGetRecSQL write FOnGetRecSQL;
    { define whether internal dataset is read only (editing is not allowed) }
    property RecDataSetReadOnly: boolean read FRecDataSetReadOnly write FRecDataSetReadOnly default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnRecDataSetEvent: TDataSetEvent read FRecDataSetEvent write SetRecDataSetEvent;
  end;

  { TrDBGrid_MS descendant with possibility to load/show/edit records from any SQL command }
  TrDBGrid_MSUDB = class(TrDBGrid_MS)
  private
    FListSQL: TStrings;
    FListConnection: TConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListDataSetReadOnly: boolean;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FOnGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh Grid content }
    procedure ReloadData;
    { return internal Dataset to allow reading of fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return internal Datasource to allow using for other DB components }
    property ListDataSource: TDataSource read FListDataSource;
  published
    { define SQL command }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadData is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FOnGetListSQL write FOnGetListSQL;
    { define whether internal dataset is read only (editing is not allowed) }
    property ListDataSetReadOnly: boolean read FListDataSetReadOnly write FListDataSetReadOnly default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { universal sorter for any UDB dataset }
  TrDBGridSorter_UDB = class(TrDBGridCustomSorter)
  protected
    function SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string): boolean; override;
    procedure DatasetRequery(DataSet: TDataSet); override;
  end;

procedure Register;

var
  lastRequeryUDB: TQueryUDB;

implementation

{$R *.res}
uses RosiCompGlobalCfg, rstring, rdlg, rtool, rdbtool;

const
  resnameDBLOOKUPDETAIL = 'DBLOOKUPDETAIL';
  resnameDBLOOKUPSELECT = 'DBLOOKUPSELECT';

type
  THackWinControl = class(TWinControl)
  end;

var
  lasttime: TTime;

procedure Register;
begin
  RegisterComponents('Rosi UDB', [TConnectionUDB, TQueryUDB, TEditUDB, TrDBEditUDB, TrDBCheckBoxUDB, TComboBoxUDB, TrDBComboBoxUDB, TrDBLookupComboBoxUDB, TListBoxUDB,
    TrDBListBoxUDB, TrDBLookupListBoxUDB, TCheckListBoxUDB, TrDBCheckListBoxUDB, TrDBRecViewUDB, TrDBGrid_MSUDB, TrDBGridSorter_UDB]);
end;

// ==============================================================================
// TConnectionUDBEx
{$IFDEF AddConReadOnly}

procedure TConnectionUDB.SetReadOnly(Value: boolean);
begin
{$IFDEF USEADO}
  if Value then
    Mode:=cmRead
  else
    Mode:=cmReadWrite;
  FReadOnly:=Value;
{$ENDIF}
{$IF Defined(USEDBX) OR Defined(USEIB) OR Defined(USEFIB) OR Defined(USEFD)}
  FReadOnly:=Value;
{$IFEND}
end;
{$ENDIF}
{$IFDEF AddConHostName}

function TConnectionUDB.GetHostName: string;
const
  ConStrIds: array [1 .. 2] of string = ('Data Source=', 'SERVER=');
var
  s: string;
  {$IFDEF USEADO} i, p: integer; {$ENDIF}
begin
{$IFDEF USEADO}
  for i:= Low(ConStrIds) to High(ConStrIds) do
  begin
    p:=AnsiPosEx(ConStrIds[i], FConnectionStringOrig);
    if p > 0 then
    begin
      s:=copy(FConnectionStringOrig, p, Length(FConnectionStringOrig));
      s:=FindStrPart(s, ';', 1, true);
      s:=copy(s, Length(ConStrIds[i]) + 1, Length(s));
      Break;
    end;
  end;
{$ENDIF}
{$IFDEF USEDBX}
  if SameText(DriverName, 'Firebird') or SameText(DriverName, 'InterBase') then
    s:=rtool.FindStrPart(Params.Values['Database'], ':', 1, false)
  else
    s:=Params.Values['HostName'];
{$ENDIF}
{$IFDEF USEIB}
  s:=rtool.FindStrPart(DatabaseName, ':', 1, false);
{$ENDIF}
{$IFDEF USEFIB}
  s:=rtool.FindStrPart(DBName, ':', 1, false);
{$ENDIF}
{$IFDEF USEFD}
  s:=Params.Values['Server'];
{$ENDIF}
  Result:=s;
end;
{$ENDIF}

function TConnectionUDB.GetDatabaseName: string;
const
  ConStrIds: array [1 .. 2] of string = ('Initial Catalog=', 'DATABASE=');
var
  s: string;
  {$IFDEF USEADO} i, p: integer; {$ENDIF}
begin
{$IFDEF USEADO}
  for i:= Low(ConStrIds) to High(ConStrIds) do
  begin
    p:=AnsiPosEx(ConStrIds[i], FConnectionStringOrig);
    if p > 0 then
    begin
      s:=copy(FConnectionStringOrig, p, Length(FConnectionStringOrig));
      s:=FindStrPart(s, ';', 1, true);
      s:=copy(s, Length(ConStrIds[i]) + 1, Length(s));
      Break;
    end;
  end;
{$ENDIF}
{$IFDEF USEDBX}
  if SameText(DriverName, 'Firebird') or SameText(DriverName, 'InterBase') then
    s:=rtool.FindStrPart(Params.Values['Database'], ':', 2, false)
  else
    s:=Params.Values['Database'];
{$ENDIF}
{$IFDEF USEIB}
  s:=rtool.FindStrPart(DatabaseName, ':', 2, false);
{$ENDIF}
{$IFDEF USEFIB}
  s:=rtool.FindStrPart(DBName, ':', 2, false);
{$ENDIF}
{$IFDEF USEFD}
  s:=Params.Values['Database'];
{$ENDIF}
  Result:=s;
end;

procedure TConnectionUDB.CallProcForRegisteredDataset(CallBackProc: TNotifyEvent);
var
  i: integer;
  Current: TDataSet;
begin
{$IF Defined(USEADO) OR Defined(USEDBX) OR Defined(USEIB) OR Defined(USEFD)}
  for i:=0 to DataSetCount - 1 do
  begin
    Current:=self.DataSets[i];
{$IFDEF USEDBX}
    if Current.Owner is TQueryUDB then
      Current:=Current.Owner as TQueryUDB;
{$ENDIF}
    try
      CallBackProc(Current);
    except
      // Ignore.
    end;
  end;
{$IFEND}
{$IFDEF USEZeos}
  for i:=0 to FDataSets.Count - 1 do
  begin
    Current:=TDataSet(FDataSets[i]);
    try
      CallBackProc(Current);
    except
      // Ignore.
    end;
  end;
{$ENDIF}
{$IFDEF USEFIB}
  for i:=0 to self.FFIBBases.Count - 1 do
  begin
    if FIBBases[i] <> nil then
      if (FIBBases[i].Owner is TDataSet) then
      begin
        Current:=TDataSet(FIBBases[i].Owner);
        try
          CallBackProc(Current);
        except
          // Ignore.
        end;
      end;
  end;
{$ENDIF}
end;
{$IFDEF USEZeos}

procedure TConnectionUDB.Open;
begin
  Connected:=true;
end;

procedure TConnectionUDB.Close;
begin
  Connected:=false;
end;
{$ENDIF}
{$IFDEF USEADO}

procedure TConnectionUDB.DoConnect;
begin
  if (csDesigning in ComponentState) and (FConnectionStringDesign <> '') then
    ConnectionString:=FConnectionStringDesign;
  FConnectionStringOrig:=ConnectionString;
  inherited;
end;
{$ENDIF}

// ==============================================================================
// TQueryUDB
constructor TQueryUDB.Create(AOwner: TComponent);
begin
  inherited;
  FFieldImgIdx:=TStringList.Create;
  FControlsToUpdate:=TStringList.Create;
  FControlsReadOnlyColor:=clBtnFace;
  FControlsUpdating:=false;
  FFieldPropsAutoLoad:=true;
  FFieldDefFormatAutoSet:=true;
  FConfirmDelete:=false;
  FFieldTagToDuplic:=1;
  FFieldTagToReadOnly:=2;
  FLastSort:='';
{$IFDEF AddQuerySQL}
  FSQL:=TStringList.Create;
  TStringList(FSQL).OnChange:=SQLChanged;
{$ENDIF}
{$IFDEF USEDBX}
  FAutoApplyUpdates:=true;
{$ENDIF}
{$IFDEF USEFIB}
  AutoCommit:=true;
  AutoUpdateOptions.AutoReWriteSqls:=true;
  AutoUpdateOptions.CanChangeSQLs:=true;
  AutoUpdateOptions.UpdateOnlyModifiedFields:=true;
{$ENDIF}
end;

destructor TQueryUDB.Destroy;
begin
  FreeAndNil(FFieldImgIdx);
  FreeAndNil(FControlsToUpdate);
{$IFDEF AddQuerySQL}
  FreeAndNil(FSQL);
{$ENDIF}
  inherited;
end;

procedure TQueryUDB.Loaded;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  lasttime:=Now;
  if FFieldDefFormatAutoSet then
    SetFieldDefaultFormat;
  if FFieldPropsAutoLoad then
    LoadFieldsProps;
  AddIntLog('init (loading field props)', false);
end;

procedure TQueryUDB.Post;
begin
  inherited;
end;

procedure TQueryUDB.AddIntLog(const act: string; AddRecCount: boolean);
var
  s, sep: string;
  C: TComponent;
begin
  if Assigned(RosiCompConfig.rDBFieldDefaultsCfg.InternalSQLLog) then
  begin
    sep:=RosiCompConfig.rDBFieldDefaultsCfg.InternalLogSeparator;
    C:=self;
    s:=C.Name;
    while C.Owner<>nil do
    begin
      C:=C.Owner;
      s:=C.Name+'.'+s;
      if (C is TForm) or (C is TDataModule) then
        Break;
    end;
    s:=TimeToStr(Now) + sep + s + sep + act + sep;
    s:=s + FormatFloat('0.0000', (Now - lasttime) * 24 * 60 * 60) + sep;
    if AddRecCount then
    begin
      s:=s + IntToStr(RecordCount) + sep;
      s:=s + IntToStr(RecordSize) + sep;
    end;
    RosiCompConfig.rDBFieldDefaultsCfg.InternalSQLLog.Add(s);
  end;
end;
{$IFDEF AddQueryReadOnly}

function TQueryUDB.GetCanModify: boolean;
begin
  Result:=not FReadOnly and inherited GetCanModify;
end;

function TQueryUDB.GetReadOnly: boolean;
begin
  Result:=FReadOnly;
end;
{$ENDIF}

procedure TQueryUDB.SetReadOnly(Value: boolean);
begin
{$IFDEF USEADO} FReadOnly:=Value; {$ENDIF}
{$IFDEF USEZeos} ReadOnly:=Value; {$ENDIF}
{$IFDEF USEDBX} ReadOnly:=Value; {$ENDIF}
{$IFDEF USEIB} FReadOnly:=Value; {$ENDIF}
{$IFDEF USEFIB} FReadOnly:=Value; {$ENDIF}
{$IFDEF USEFD} FReadOnly:=Value; {$ENDIF}
  UpdateDatasetState;
end;
{$IFDEF AddQuerySQL}

function TQueryUDB.GetSQL: TStrings;
begin
{$IFDEF USEDBX}
  if Assigned(DataSet) then
    FSQL.Text:=DataSet.CommandText;
{$ENDIF}
{$IFDEF USEFIB}
  if Assigned(SQLs.SelectSQL) then
    FSQL.Text:=SQLs.SelectSQL.Text;
{$ENDIF}
  Result:=FSQL;
end;

procedure TQueryUDB.SetSQL(Value: TStrings);
begin
  FSQL.Assign(Value);
{$IFDEF USEDBX}
  if Assigned(DataSet) then
    DataSet.CommandText:=FSQL.Text;
{$ENDIF}
{$IFDEF USEFIB}
  if Assigned(SQLs.SelectSQL) then
    SQLs.SelectSQL.Text:=FSQL.Text;
{$ENDIF}
end;

procedure TQueryUDB.SQLChanged(Sender: TObject);
begin
{$IFDEF USEDBX}
  if Assigned(DataSet) and (DataSet.CommandText <> FSQL.Text) then
    DataSet.CommandText:=FSQL.Text;
{$ENDIF}
{$IFDEF USEFIB}
  if Assigned(SQLs.SelectSQL) and (SQLs.SelectSQL.Text <> FSQL.Text) then
    SQLs.SelectSQL.Text:=FSQL.Text;
{$ENDIF}
end;
{$ENDIF}
{$IFDEF AddQueryConnection}

function TQueryUDB.GetConnection: {$IFDEF USEIB} TIBDataBase {$ENDIF} {$IFDEF USEFIB} TFIBDatabase {$ENDIF};
begin
  Result:=Database;
end;

procedure TQueryUDB.SetConnection(Value: {$IFDEF USEIB} TIBDataBase {$ENDIF} {$IFDEF USEFIB} TFIBDatabase {$ENDIF});
begin
  Database:=Value;
end;
{$ENDIF}
{$IFDEF USEDBX}

procedure TQueryUDB.IntApplyUpdates;
begin
  if FAutoApplyUpdates then
  begin
    if not Assigned(OnReconcileError) then
      OnReconcileError:=IntReconcileError;
    ApplyUpdates(0);
  end;
end;

procedure TQueryUDB.IntReconcileError(DataSet: TCustomClientDataSet; E: EReconcileError; UpdateKind: TUpdateKind; var Action: TReconcileAction);
begin
  ShowErr(rstring.sSaveDataErr, E);
  Action:=raCancel;
  // RequeryEx;
end;
{$ENDIF}

procedure TQueryUDB.DoBeforeOpen;
var
  ro: boolean;
begin
  if not(csDesigning in ComponentState) then
  begin
    if Connection is TConnectionUDB then
      ro:=TConnectionUDB(Connection).FReadOnly
    else
      ro:=false;
    if ro and not ReadOnly then
      ReadOnly:=true;
  end;
  inherited;
  CallDatasetEvent(deBeforeOpen);
end;

procedure TQueryUDB.DoAfterOpen;
begin
  if FLastSort <> '' then
    try
{$IFDEF USEADO} Sort:=FLastSort; {$ENDIF}
{$IFDEF USEZeos} SortedFields:=FLastSort; {$ENDIF}
{$IFDEF USEDBX} IndexName:=FLastSort; {$ENDIF}
    except
    end;
{$IFDEF USEFIB}
  try
    if FAutoUpdateOptions.UpdateTableName = '' then
      FAutoUpdateOptions.UpdateTableName:=GetTableName;
    if FAutoUpdateOptions.KeyFields = '' then
      FAutoUpdateOptions.KeyFields:=IdFieldName;
  except
  end;
{$ENDIF}
{$IFDEF USEFD} IndexName:=FLastSort; {$ENDIF}
  inherited;
  CallDatasetEvent(deAfterOpen);
end;

procedure TQueryUDB.DoBeforeClose;
begin
  if Active then
  begin
{$IFDEF USEADO} FLastSort:=Sort; {$ENDIF}
{$IFDEF USEZeos} FLastSort:=SortedFields; {$ENDIF}
{$IFDEF USEDBX} FLastSort:=IndexName; {$ENDIF}
{$IFDEF USEFD} FLastSort:=IndexName; {$ENDIF}
  end;
  inherited;
  CallDatasetEvent(deBeforeClose);
end;

procedure TQueryUDB.DoAfterClose;
begin
  inherited;
  CallDatasetEvent(deAfterClose);
end;

procedure TQueryUDB.DoBeforePost;
var
  i: integer;
  F: TField;
begin
  if FCheckFieldSizeBeforePost then
    for i:=0 to self.FieldCount - 1 do
    begin
      F:=self.Fields[i];
      if (F is TStringField) and (Length(F.AsString) > F.Size) then
        F.AsString:=copy(F.AsString, 1, F.Size);
    end;
  inherited;
  CallDatasetEvent(deBeforePost);
end;

procedure TQueryUDB.DoAfterPost;
begin
  inherited;
  CallDatasetEvent(deAfterPost);
{$IFDEF USEDBX}
  IntApplyUpdates;
{$ELSE}
  CallDatasetEvent(deAfterPostOrApplyUpdates);
{$ENDIF}
end;

procedure TQueryUDB.DoBeforeScroll;
begin
  inherited;
  CallDatasetEvent(deBeforeScroll);
end;

procedure TQueryUDB.DoAfterScroll;
begin
  inherited;
  CallDatasetEvent(deAfterScroll);
end;

procedure TQueryUDB.DoBeforeCancel;
begin
  inherited;
  CallDatasetEvent(deBeforeCancel);
end;

procedure TQueryUDB.DoAfterCancel;
begin
  inherited;
  CallDatasetEvent(deAfterCancel);
end;

procedure TQueryUDB.DoBeforeInsert;
begin
  inherited;
  CallDatasetEvent(deBeforeInsert);
end;

procedure TQueryUDB.DoAfterInsert;
begin
  inherited;
  CallDatasetEvent(deAfterInsert);
end;

procedure TQueryUDB.DoBeforeEdit;
begin
  inherited;
  CallDatasetEvent(deBeforeEdit);
end;

procedure TQueryUDB.DoAfterEdit;
begin
  inherited;
  CallDatasetEvent(deAfterEdit);
end;

procedure TQueryUDB.DoBeforeDelete;
begin
  inherited;
  if FConfirmDelete then
    if DlgQ(rstring.sDelRecordQ) <> mrYes then
      Abort;
  CallDatasetEvent(deBeforeDelete);
end;

procedure TQueryUDB.DoAfterDelete;
begin
  inherited;
  CallDatasetEvent(deAfterDelete);
{$IFDEF USEDBX}
  IntApplyUpdates;
{$ELSE}
  CallDatasetEvent(deAfterPostOrApplyUpdates);
{$ENDIF}
end;
{$IFDEF USEDBX}

procedure TQueryUDB.DoBeforeApplyUpdates(var OwnerData: OleVariant);
begin
  inherited;
  CallDatasetEvent(deBeforeApplyUpdates);
end;

procedure TQueryUDB.DoAfterApplyUpdates(var OwnerData: OleVariant);
begin
  inherited;
  CallDatasetEvent(deAfterApplyUpdates);
  CallDatasetEvent(deAfterPostOrApplyUpdates);
end;
{$ENDIF}

procedure TQueryUDB.CallDatasetEvent(Event: TDatasetEventType);
var
  i: integer;
begin
  for i:=0 to High(FEventNotifyArr) do
  begin
    if @FEventNotifyArr[i] <> nil then
      FEventNotifyArr[i](self, Event);
  end;
  if Assigned(FDataSetEvent) then
    FDataSetEvent(self, Event);

  if Event in [deAfterOpen, deAfterPost, deAfterCancel, deAfterInsert, deAfterEdit, deAfterDelete, deAfterPostOrApplyUpdates] then
    ProcControlsToUpdate(Event);
end;

procedure TQueryUDB.ProcControlsToUpdate(Event: TDatasetEventType);
var
  i: integer;
  C: TComponent;
  W: TWinControl;
  v: string;
begin
  if Owner is TComponent then
  begin
    C:=Owner as TComponent;
    for i:=0 to C.ComponentCount - 1 do
      if C.Components[i] is TWinControl then
      begin
        W:=C.Components[i] as TWinControl;
        v:=FControlsToUpdate.Values[W.Name];
        if (v <> '') then
        begin
          if AnsiPosEx('EN', v) > 0 then
            W.Enabled:=IsEditing;
          if AnsiPosEx('RO', v) > 0 then
          begin
            if W is TCustomEdit then
              TCustomEdit(W).ReadOnly:=not IsEditing;
          end;
          if AnsiPosEx('COL', v) > 0 then
          begin
            if IsEditing then
              THackWinControl(W).Color:=clWindow
            else
              THackWinControl(W).Color:=FControlsReadOnlyColor;
          end;
          if AnsiPosEx('EDIT', v) > 0 then
          begin
            if W is TEdit then
              TEdit(W).OnChange:=ControlsToUpdateChange;
            if W is TMemo then
              TMemo(W).OnChange:=ControlsToUpdateChange;
            if W is TComboBox then
              TComboBox(W).OnChange:=ControlsToUpdateChange;
            if W is TListBox then
              TListBox(W).OnClick:=ControlsToUpdateChange;
            if W is TCheckBox then
              TCheckBox(W).OnClick:=ControlsToUpdateChange;
            if W is TRadioGroup then
              TRadioGroup(W).OnClick:=ControlsToUpdateChange;
          end;
          if (AnsiPosEx('RESET', v) > 0) and not IsEditing then
          begin
            FControlsUpdating:=true;
            try
              if W is TCustomEdit then
                TCustomEdit(W).Text:='';
            finally
              FControlsUpdating:=false;
            end;
          end;
        end;
      end;
  end;
end;

procedure TQueryUDB.ControlsToUpdateChange(Sender: TObject);
begin
  if FControlsUpdating or IsEditing then
    Exit;
  if GetCanModify then
    Edit;
end;

procedure TQueryUDB.SetSortEx(const ASort: string);
var
  B: TBookmark;
begin
  FLastSort:=ASort;
  if Active then
  begin
    B:=GetBookmark;
    try
{$IFDEF USEADO} Sort:=ASort; {$ENDIF}
{$IFDEF USEZeos} SortedFields:=ASort; {$ENDIF}
{$IFDEF USEDBX} IndexName:=ASort; {$ENDIF}
{$IFDEF USEFD} IndexName:=ASort; {$ENDIF}
      GotoBookmark(B);
    finally
      FreeBookmark(B);
    end;
  end;
end;

procedure TQueryUDB.SetFieldImgIdx(Value: TStrings);
begin
  FFieldImgIdx.Assign(Value);
end;

procedure TQueryUDB.SetControlsToUpdate(Value: TStrings);
begin
  FControlsToUpdate.Assign(Value);
end;

function TQueryUDB.HasFieldToDuplic: boolean;
var
  i: integer;
begin
  Result:=false;
  if FFieldTagToDuplic = 0 then
    Exit;
  for i:=0 to FieldCount - 1 do
    if (Fields[i].Tag and FFieldTagToDuplic) > 0 then
    begin
      Result:=true;
      Exit;
    end;
end;

function TQueryUDB.SetFieldsReadOnly: integer;
var
  i: integer;
begin
  Result:=0;
  if FFieldTagToReadOnly = 0 then
    Exit;
  for i:=0 to FieldCount - 1 do
  begin
    Fields[i].ReadOnly:=(Fields[i].Tag and FFieldTagToReadOnly) > 0;
    if Fields[i].ReadOnly then
      Inc(Result);
  end;
end;

function TQueryUDB.DuplicateRecord: boolean;
begin
  Result:=FFieldTagToDuplic > 0;
  if Result then
    rdbtool.DuplicRecord(self, FFieldTagToDuplic, false);
end;

procedure TQueryUDB.AddEventNotify(CallBackProc: TDataSetEvent);
var
  i, max: integer;
begin
  max:=Length(FEventNotifyArr);
  { for I:=0 to max - 1 do
    if @FEventNotifyArr[I] = Addr(CallBackProc)^ then
    Exit;
    }
  for i:=0 to max - 1 do
    if @FEventNotifyArr[i] = nil then
    begin
      FEventNotifyArr[i]:=CallBackProc;
      Exit;
    end;
  SetLength(FEventNotifyArr, max + 1);
  FEventNotifyArr[max]:=CallBackProc;
end;

procedure TQueryUDB.RemoveEventNotify(CallBackProc: TDataSetEvent);
var
  i, max: integer;
begin
  max:=Length(FEventNotifyArr);
  for i:=0 to max - 1 do
    if @FEventNotifyArr[i] = @CallBackProc then
      FEventNotifyArr[i]:=nil;
end;

procedure TQueryUDB.RequeryEx(ChangeCursor: boolean = true);
var
  B: TBookmark;
  oldCursor: TCursor;
  id: variant;
begin
  lastRequeryUDB:=self;
  if ChangeCursor then
    oldCursor:=ChCursor(RosiCompConfig.rDBComponentsCfg.RequeryCursorType)
  else
    oldCursor:=Screen.Cursor;
  try
    try
      lasttime:=Now;
      if Active then
      begin
        id:=Fields[0].AsVariant;
        B:=GetBookmark;
        DisableControls;
        try
  {$IFDEF USEADO} Requery; {$ENDIF}
  {$IFDEF USEZeos} Close; Open; {$ENDIF}// BUG - RereadRows do not work, must Close and Open
  {$IFDEF USEDBX} Close; Open; {$ENDIF}
  {$IFDEF USEIB} Requery; {$ENDIF}
  {$IFDEF USEFIB} Close; Open; {$ENDIF}
  {$IFDEF USEFD} Close; Open; {$ENDIF}
          try
            GotoBookmark(B);
          except
            try
              if not VarIsNull(id) then
                Locate(Fields[0].FieldName, id, []);
            except
            end;
          end;
        finally
          EnableControls;
        end;
      end
      else
        Open;

      if ChangeCursor then
        AddIntLog('opening (W)', true)
      else
        AddIntLog('opening', true);
    except
      AddIntLog(' ERROR opening ' + rtool.ReplaceStr(#13#10, ' ', self.SQL.Text), false);
      raise;
    end;
  finally
    if ChangeCursor then
      ChCursor(oldCursor);
  end;
end;

procedure TQueryUDB.SetSQLWhere(const Condition: string);
begin
  rdbtool.SetSQLWhere(SQL, Condition);
end;

procedure TQueryUDB.SetSQLWhere(const FormatCondition: string; Value: variant);
begin
  rdbtool.SetSQLWhere(SQL, FormatCondition, Value);
end;

procedure TQueryUDB.SetSQLWhere(Field: TField);
begin
  rdbtool.SetSQLWhere(SQL, Field);
end;

procedure TQueryUDB.SetSQLWhere(Field: TField; Value: variant);
begin
  rdbtool.SetSQLWhere(SQL, Field, Value);
end;

procedure TQueryUDB.AddSQLWhereOR(const Condition: string);
begin
  rdbtool.AddSQLWhereOR(SQL, Condition);
end;

procedure TQueryUDB.AddSQLWhereAND(const Condition: string);
begin
  rdbtool.AddSQLWhereAND(SQL, Condition);
end;

function TQueryUDB.SetExecSQL(const SQLCmd: string): integer;
begin
  Close;
  SQL.Text:=SQLCmd;
{$IFDEF USEADO} Result:=ExecSQL; {$ENDIF}
{$IFDEF USEZeos} Result:=0; ExecSQL; {$ENDIF}
{$IFDEF USEDBX} Result:=0; Execute; {$ENDIF}
{$IFDEF USEIB} Result:=0; ExecSQL; {$ENDIF}
{$IFDEF USEFIB} Result:=0; Open; {$ENDIF}
{$IFDEF USEFD} Result:=0; ExecSQL; {$ENDIF}
  Close;
end;

procedure TQueryUDB.SetFieldDefaultFormat;
var
  i: integer;
  F: TField;
begin
  for i:=0 to self.FieldCount - 1 do
  begin
    F:=self.Fields[i];
    if F is TFloatField then
    begin
      if TFloatField(F).DisplayFormat = '' then
        TFloatField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultFloatDisplayFormat;
      if TFloatField(F).EditFormat = '' then
        TFloatField(F).EditFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultFloatEditFormat;
    end;
    { if F is TBCDField then
      begin
      if TBCDField(F).DisplayFormat='' then
        TBCDField(F).DisplayFormat:=DefaultFloatDisplayFormat;
      if TBCDField(F).EditFormat='' then
        TBCDField(F).EditFormat:=DefaultFloatEditFormat;
      end;
      }
    if F is TBooleanField then
    begin
      if TBooleanField(F).DisplayValues = '' then
        TBooleanField(F).DisplayValues:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultBoolDisplayFormat;
    end;
    if F is TDateTimeField then
    begin
      if TDateTimeField(F).DisplayFormat = '' then
      begin
        TDateTimeField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultDateTimeDisplayFormat;
        if SameText(copy(F.FieldName, 1, 4), 'Time') then
          TDateTimeField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultTimeDisplayFormat;
      end;
    end;
  end;
end;

procedure TQueryUDB.LoadFieldsProps;
begin
  if Assigned(RosiCompConfig.rDBFieldDefaultsCfg.LoadFieldsPropsProc) then
    RosiCompConfig.rDBFieldDefaultsCfg.LoadFieldsPropsProc(self);
end;
{$IFDEF USEZeos}

function TQueryUDB.FindMetadataColumnIndex(const FieldName: string): integer;
var
  i: integer;
begin
  Result:=-1;
  if Assigned(ResultSet) and Assigned(ResultSet.GetMetadata) then
  begin
    for i:=1 to ResultSet.GetMetadata.GetColumnCount do
      if SameText(ResultSet.GetMetadata.GetColumnName(i), FieldName) then
      begin
        Result:=i; Break;
      end;
  end;
end;
{$ENDIF}

function TQueryUDB.GetTableName: string;
begin
  Result:=self.TableName;
  if Result = '' then
    Result:=rdbtool.GetSQLTableName(self.SQL);
end;

procedure TQueryUDB.UpdateDatasetState;
begin
  DataEvent(deDataSetChange, 0);
end;

function TQueryUDB.IsEditing: boolean;
begin
  Result:=(State in [dsEdit, dsInsert]);
end;

function TQueryUDB.IsNewRec: boolean;
begin
  Result:=(State in [dsInsert]);
end;

constructor TQueryIntList.Create(AOwner: TComponent);
begin
  inherited;
  FFieldPropsAutoLoad:=false;
  FFieldDefFormatAutoSet:=false;
end;

// ==============================================================================
// TBoundComboButton

constructor TBoundComboButton.CreateBtn(AOwner: TComponent; AName, ACaption, AGlyphResourceName: string; Idx: integer);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  SetSubComponent(true);
  if AOwner is TControl then
    ParentW:=AOwner as TControl;
  if AName = '' then
    AName:='ComboButton';
  Name:=AName;
  Caption:='';
  FIdx:=Idx;
  GlyphResourceName:=AGlyphResourceName;
  if GlyphResourceName <> '' then
  begin
{$IFDEF DoNotUseGlyphInDesign}
    if (csDesigning in ComponentState) then
      Caption:=IntToStr(FIdx)
    else
{$ENDIF}
    begin
      Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex(GlyphResourceName), FScaleRatio));
    end;
  end
  else
    Caption:=ACaption;
end;

procedure TBoundComboButton.Loaded;
begin
  inherited;
{$IFDEF DoNotUseGlyphInDesign}
    if (csDesigning in ComponentState) then
      Glyph.Width:=0;
{$ENDIF}
  if Glyph.Width > 0 then
    Caption:='';
end;

procedure TBoundComboButton.ChangeScale(M: integer; D: integer{$IFDEF DX102P}; isDpiChange: Boolean{$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  if not (csDesigning in ComponentState) then
    Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex(GlyphResourceName), FScaleRatio));
end;

procedure TBoundComboButton.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  RefreshPosition;
end;

procedure TBoundComboButton.SetIdx(Value: integer);
begin
  FIdx:=Value;
  RefreshPosition;
end;

procedure TBoundComboButton.RefreshPosition;
var
  Size: integer;
begin
  if Assigned(ParentW) then
  begin
    Size:=ParentW.Height - 2;
    if Size > 24 then
      Size:=24;
    if Visible then
      SetBounds(ParentW.Left + ParentW.Width + 2 + (FIdx - 1) * Size, ParentW.Top + 1, Size, Size)
    else
      SetBounds(ParentW.Left + ParentW.Width - Size, ParentW.Top + 1, Size, Size);
  end;
end;

procedure TBoundComboButton.Click;
begin
  if Assigned(Owner) and (Owner is TWinControl) then
    try
      TWinControl(Owner).SetFocus;
    except
    end;
  inherited;
end;

// ==============================================================================
// TCheckBox

constructor TrDBCheckBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  self.ValueChecked:='1';
  self.ValueUnchecked:='0';
end;

// ==============================================================================
// TEdit

constructor TEditUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
  Color:=clBtnFace;
end;

destructor TEditUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TEditUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TEditUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TEditUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TEditUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TEditUDB.CMHintShow(var Message: TMessage);
var
  hnt: string;
  ACanvas: TControlCanvas;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    ACanvas:=TControlCanvas.Create;
    try
      ACanvas.Control:=self;
      ACanvas.Font.Assign(Font);
      if ACanvas.TextWidth(hnt) > self.ClientWidth - 4 then
        TCMHintShow(Message).HintInfo^.HintStr:=hnt;
    finally
      FreeAndNil(ACanvas);
    end;
  end;
  inherited;
end;

procedure TEditUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
  ReadOnly:=true;
end;

procedure TEditUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TEditUDB.ReloadItems;
var
  F: TField;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      F:=FListDataSet.Fields[0];
      if IsMemoField(F) then
        self.Text:=F.AsString
      else
        self.Text:=F.DisplayText;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TComboBox

constructor TComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TComboBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TComboBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TComboBoxUDB.CMHintShow(var Message: TMessage);
var
  hnt: string;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    if Canvas.TextWidth(hnt) > self.ClientWidth - 24 then
      TCMHintShow(Message).HintInfo^.HintStr:=hnt;
  end;
  inherited;
end;

procedure TComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TComboBoxUDB.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if FDelKeyForNullValue and (Style = csDropDownList) and (Key = VK_DELETE) then
  begin
    ItemIndex:=-1;
    Click;
  end;
end;

procedure TComboBoxUDB.ReloadItems;
var
  id: integer;
  s: string;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);
          Items.AddObject(s, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TComboBoxUDB.GetListField(FieldIndex: integer): TField;
var
  i: integer;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  FListDataSet.First;
  for i:=0 to ItemIndex - 1 do
    FListDataSet.Next;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TComboBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TComboBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TrDBEditUDB

constructor TrDBEditUDB.Create(AOwner: TComponent);
begin
  inherited;
  FEditButton:=TBoundComboButton.CreateBtn(self, '', '...', '', 1);
  FEditButton.FreeNotification(self);
  //FEditButton.OnClick:=ComboButtonClick;
  FEditButtonVisible:=true;
  FEditButtonAllwaysEnabled:=true;
end;

destructor TrDBEditUDB.Destroy;
begin
  FreeAndNil(FEditButton);
  inherited;
end;

procedure TrDBEditUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FEditButton) and (Operation = opRemove) then
    FEditButton:=nil;
end;

procedure TrDBEditUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FEditButton) then
    FEditButton.RefreshPosition;
end;

procedure TrDBEditUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FEditButton) then
    FEditButton.Parent:=AParent;
end;

procedure TrDBEditUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FEditButton) then
    FEditButton.Enabled:=Enabled;
end;

procedure TrDBEditUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FEditButton) then
    FEditButton.Visible:=Visible and FEditButtonVisible;
end;

procedure TrDBEditUDB.SetEditButtonVisible(Value: boolean);
begin
  FEditButtonVisible:=Value;
  if Assigned(FEditButton) then
    FEditButton.Visible:=Visible and FEditButtonVisible;
end;

procedure TrDBEditUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FEditButton) then
    FEditButton.Enabled:=Enabled and not ro and (FEditButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

// ==============================================================================
// TrDBComboBox

constructor TrDBComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TrDBComboBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TrDBComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TrDBComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TrDBComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TrDBComboBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TrDBComboBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TrDBComboBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TrDBComboBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TrDBComboBoxUDB.GetListConnection: TConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TConnectionUDB(TQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TrDBComboBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TrDBComboBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBComboBoxUDB.ComboButtonClick(Sender: TObject);
var
  tbl: string;
  F: TField;
begin
  tbl:=ListButtonTableName;
  if tbl = '' then
    tbl:=rdbtool.GetSQLTableName(ListSQL);
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, tbl, Field, F);
end;

procedure TrDBComboBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TrDBComboBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
        TQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TrDBComboBoxUDB.ReloadItems;
var
  id: integer;
  s: string;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if (FListDataSet.FieldCount > 1) and (FListDataSet.Fields[1] is TNumericField) then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);
          Items.AddObject(s, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
    if Assigned(Field) and Field.DataSet.Active then
      try
        ItemIndex:=Items.IndexOf(Field.AsString);
      except
      end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TrDBComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TrDBComboBoxUDB.GetListField(FieldIndex: integer): TField;
var F: TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  F:=FDataLink.DataSet.FindField(DataField);
  if Assigned(F) then
    rdbtool.LocateRecord(FListDataSet.Fields[0],F.AsString);
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TrDBLookupComboBox

constructor TrDBLookupComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListDataSource:=TDataSource.Create(self);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FListDetailButton:=TBoundComboButton.CreateBtn(self, 'ComboDetailButton', '', resnameDBLOOKUPDETAIL, 2);
  FListDetailButton.FreeNotification(self);
  FListDetailButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
  FListDetailButtonVisible:=true;
end;

destructor TrDBLookupComboBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FListDataSource.DataSet:=nil;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListDataSource);
  FreeAndNil(FListButton);
  FreeAndNil(FListDetailButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TrDBLookupComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TrDBLookupComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
  if (AComponent = FListDetailButton) and (Operation = opRemove) then
    FListDetailButton:=nil;
end;

procedure TrDBLookupComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

function TrDBLookupComboBoxUDB.GetListDetailButtonEnability: boolean;
begin
  Result:=Enabled and not VarIsNull(KeyValue) and (not Assigned(Field) or not Field.IsNull);
end;

procedure TrDBLookupComboBoxUDB.KeyValueChanged;
begin
  inherited;
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TrDBLookupComboBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
  if Assigned(FListDetailButton) then
    FListDetailButton.RefreshPosition;
end;

procedure TrDBLookupComboBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
  if Assigned(FListDetailButton) then
    FListDetailButton.Parent:=AParent;
end;

procedure TrDBLookupComboBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TrDBLookupComboBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
  if Assigned(FListDetailButton) then
    FListDetailButton.Visible:=Visible and FListDetailButtonVisible;
end;

procedure TrDBLookupComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TrDBLookupComboBoxUDB.GetListConnection: TConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TConnectionUDB(TQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TrDBLookupComboBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBLookupComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TrDBLookupComboBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
  if Assigned(FListDetailButton) then
    if FListButtonVisible then
      FListDetailButton.Idx:=2
    else
      FListDetailButton.Idx:=1;
end;

procedure TrDBLookupComboBoxUDB.SetListDetailButtonVisible(Value: boolean);
begin
  FListDetailButtonVisible:=Value;
  if Assigned(FListDetailButton) then
    FListDetailButton.Visible:=Visible and FListDetailButtonVisible;
end;

procedure TrDBLookupComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBLookupComboBoxUDB.ComboButtonClick(Sender: TObject);
var
  tbl: string;
  F: TField;
begin
  tbl:=ListButtonTableName;
  if tbl = '' then
    tbl:=rdbtool.GetSQLTableName(ListSQL);
  F:=nil;
  if Assigned(ListSource) and Assigned(ListSource.DataSet) and ListSource.DataSet.Active then
    F:=ListSource.DataSet.FieldByName(KeyField);
  if (Sender = FListButton) and Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, tbl, Field, F);
  if (Sender = FListDetailButton) and Assigned(FClickComboDetailButton) then
    FClickComboDetailButton(self, GetListConnection, tbl, Field, F);
end;

procedure TrDBLookupComboBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TrDBLookupComboBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  // reload items after reconnecting
  try
    if FListAutoLoad and (FListDataSource.DataSet=FListDataSet) and not (FListDataSet.Active) and (FListDataSet.Connection.Connected) then
      ReloadItems;
  except
  end;

  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TrDBLookupComboBoxUDB.ReloadItems;
begin
  try
    if (FListSQL.Count = 0) then
    begin
      // standard ListSource component is used
      if Assigned(ListSource) then
      begin
        ListSource.DataSet.Close;
        ListSource.DataSet.Open;
      end;
      Exit;
    end;
    // internal SQL is used
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      FListDataSet.LoadFieldsProps;
      FListDataSource.DataSet:=FListDataSet;
      ListSource:=FListDataSource;
      if KeyField='' then
        KeyField:=FListDataSet.Fields[0].FieldName;
      if ListField='' then
        ListField:=FListDataSet.Fields[1].FieldName;
      DBEditLabel.SetLabelField(ListSource, ListField);
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TrDBLookupComboBoxUDB.GetListValue: string;
var
  F: TField;
begin
  F:=FListDataSet.FindField(ListField);
  if Assigned(F) and not VarIsNull(KeyValue) then
    Result:=F.AsString
  else
    Result:='';
end;

function TrDBLookupComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TrDBLookupComboBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TrDBLookupComboBoxUDB.ListActive: boolean;
begin
  Result:=Assigned(FListDataSet) and FListDataSet.Active;
end;

// ==============================================================================
// TListBox

constructor TListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TListBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TListBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TListBoxUDB.ReloadItems;
var
  id: integer;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TrDBListBox

constructor TrDBListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TrDBListBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TrDBListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TrDBListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TrDBListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TrDBListBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TrDBListBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TrDBListBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TrDBListBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TrDBListBoxUDB.GetListConnection: TConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TConnectionUDB(TQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TrDBListBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TrDBListBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBListBoxUDB.ComboButtonClick(Sender: TObject);
var
  F: TField;
begin
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, ListButtonTableName, self.Field, F);
end;

procedure TrDBListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TrDBListBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TrDBListBoxUDB.ReloadItems;
var
  id, MaxLength: integer;
  s: string;
  F: TField;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;

          // limit length of text according field size
          MaxLength:=0;
          if Assigned(DataSource) and Assigned(DataSource.DataSet) then
          begin
            F:=DataSource.DataSet.FindField(DataField);
            if Assigned(F) and (F.DataType in [ftString, ftWideString]) and (MaxLength = 0) then
              MaxLength:=F.Size;
          end;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);

          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id)); FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
    if Assigned(Field) and Field.DataSet.Active then
      try
        ItemIndex:=Items.IndexOf(Field.AsString);
      except
      end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TrDBListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TrDBListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TrDBLookupListBox

constructor TrDBLookupListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListDataSource:=TDataSource.Create(self);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TrDBLookupListBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FListDataSource.DataSet:=nil;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListDataSource);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TrDBLookupListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TrDBLookupListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TrDBLookupListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TrDBLookupListBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TrDBLookupListBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TrDBLookupListBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TrDBLookupListBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBLookupListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TrDBLookupListBoxUDB.GetListConnection: TConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TConnectionUDB(TQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TrDBLookupListBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBLookupListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TrDBLookupListBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TrDBLookupListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBLookupListBoxUDB.ComboButtonClick(Sender: TObject);
var
  F: TField;
begin
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, ListButtonTableName, Field, F);
end;

procedure TrDBLookupListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TrDBLookupListBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TrDBLookupListBoxUDB.ReloadItems;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    FListDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FListDataSource.DataSet:=FListDataSet;
      ListSource:=FListDataSource;
      KeyField:=FListDataSet.Fields[0].FieldName;
      ListField:=FListDataSet.Fields[1].FieldName;
      DBEditLabel.SetLabelField(ListSource, ListField);
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TrDBLookupListBoxUDB.GetListValue: string;
var
  F: TField;
begin
  F:=FListDataSet.FindField(ListField);
  if Assigned(F) and not VarIsNull(KeyValue) then
    Result:=F.AsString
  else
    Result:='';
end;

function TrDBLookupListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TrDBLookupListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TCheckListBoxUDB

constructor TCheckListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TCheckListBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TCheckListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCheckListBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCheckListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCheckListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCheckListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCheckListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCheckListBoxUDB.ReloadItems;
var
  id: integer;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCheckListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCheckListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TCheckListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TCheckListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TrDBCheckListBoxUDB

constructor TrDBCheckListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FItemsListDataSet:=TQueryIntList.Create(self);
  FItemsListSQL:=TStringList.Create;
  FItemsListAutoLoad:=true;
  FItemsValueDataSet:=TQueryIntList.Create(self);
  FItemsValueSQL:=TStringList.Create;
  FLinkedDSEvent:=TList.Create;
end;

destructor TrDBCheckListBoxUDB.Destroy;
var
  i: integer;
begin
  FItemsListDataSet.Close;
  FreeAndNil(FItemsListDataSet);
  FItemsValueDataSet.Close;
  FreeAndNil(FItemsValueDataSet);
  FreeAndNil(FItemsListSQL);
  FreeAndNil(FItemsValueSQL);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TrDBCheckListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FItemsListAutoLoad then
      ReloadItemsList(false);
end;

procedure TrDBCheckListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

function TrDBCheckListBoxUDB.GetListConnection: TConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TConnectionUDB(TQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TrDBCheckListBoxUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBCheckListBoxUDB.SetItemsListSQL(Value: TStrings);
begin
  FItemsListSQL.Assign(Value);
end;

procedure TrDBCheckListBoxUDB.SetItemsListSQLTryNow(Value: boolean);
begin
  FItemsListSQLTryNow:=Value;
  try
    if FItemsListSQLTryNow then
      ReloadItemsList(false);
  finally
    FItemsListSQLTryNow:=false;
  end;
end;

procedure TrDBCheckListBoxUDB.SetItemsValueSQL(Value: TStrings);
begin
  FItemsValueSQL.Assign(Value);
end;

procedure TrDBCheckListBoxUDB.SetItemsValueSQLTryNow(Value: boolean);
begin
  FItemsValueSQLTryNow:=Value;
  try
    if FItemsValueSQLTryNow then
      ReloadItemsValue;
  finally
    FItemsValueSQLTryNow:=false;
  end;
end;

procedure TrDBCheckListBoxUDB.SetItemsValueDataSetEvent(Value: TDataSetEvent);
begin
  FItemsValueDataSetEvent:=Value;
  if Assigned(FItemsValueDataSet) then
    FItemsValueDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBCheckListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
    case EventType of
      deAfterPostOrApplyUpdates:
        if PostponedSaving then
          DoSaveItemValues;
      deAfterCancel:
        DoCancelItemValues;
    end;
end;

procedure TrDBCheckListBoxUDB.ReloadItemsList(ReloadValues: boolean);
begin
  try
    FItemsListDataSet.Close;
    FItemsListDataSet.Connection:=GetListConnection;
    if Assigned(FGetItemsListSQL) then
      FGetItemsListSQL(self, ItemsListSQL);
    FItemsListDataSet.SQL.Assign(ItemsListSQL);
    FItemsListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // FItemsListDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FAllItems.Clear;
      while not FItemsListDataSet.Eof do
      begin
        FAllItems.AddCheckItem(FItemsListDataSet.Fields[0].AsInteger, FItemsListDataSet.Fields[1].AsString, false);
        FItemsListDataSet.Next;
      end;
    end
    else
    begin
      FItemsListDataSet.Close;
      Beep;
    end;
    if ReloadValues then
      DoLoadItemValues;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

procedure TrDBCheckListBoxUDB.ReloadItemsValue;
begin
  try
    FItemsValueDataSet.Close;
    FItemsValueDataSet.Connection:=GetListConnection;
    // add notify link
    if Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TQueryUDB) then
      if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
      begin
        FLinkedDSEvent.Add(DataSource.DataSet);
        TQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
      end;

    if Assigned(FDataLink.Field) then
    begin
      // add where condition if not used
      if rdbtool.GetSQLWhere(FItemsValueSQL) = '' then
        FItemsValueSQL.Add('WHERE 1=0');
      // set where cond according master field
      rdbtool.SetSQLWhere(FItemsValueSQL, DataField + '=' + IntToStr(FDataLink.Field.AsInteger));
    end;
    if Assigned(FGetItemsValueSQL) then
      FGetItemsValueSQL(self, FItemsValueSQL);
    FItemsValueDataSet.SQL.Assign(FItemsValueSQL);
    FItemsValueDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // FItemsValueDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FAllItems.ResetItemsChecked;
      FItemsValueDataSet.First;
      while not FItemsValueDataSet.Eof do
      begin
        FAllItems.SetItemChecked(FItemsValueDataSet.Fields[1].AsInteger, true);
        FItemsValueDataSet.Next;
      end;
      FilterItemsList;
    end
    else
    begin
      FItemsValueDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

procedure TrDBCheckListBoxUDB.DoLoadItemValues;
begin
  if not PostponedSaving then
    ReloadItemsValue;
  inherited;
end;

procedure TrDBCheckListBoxUDB.DoSaveItemValues;
var
  i, id: integer;
begin
  if Assigned(FOnBeforeSaveItemValues) then
    FOnBeforeSaveItemValues(Self, FAllItems);
  for i:=0 to FAllItems.Count - 1 do
  begin
    id:=FAllItems.CheckItem(i).id;
    if FAllItems.CheckItem(i).Checked then
    begin
      if not LocateRecord(FItemsValueDataSet.Fields[1], id) then
      begin
        FItemsValueDataSet.Append;
        FItemsValueDataSet.Fields[0].AsInteger:=FDataLink.Field.AsInteger;
        FItemsValueDataSet.Fields[1].AsInteger:=id;
        FItemsValueDataSet.Post;
      end;
    end
    else
    begin
      if LocateRecord(FItemsValueDataSet.Fields[1], id) then
        FItemsValueDataSet.Delete;
    end;
  end;
  inherited;
end;

function TrDBCheckListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FItemsListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TrDBCheckListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FItemsListDataSet) or not FItemsListDataSet.Active then
    DoLoadItemList;
  if not FItemsListDataSet.Active then
    Exit;
  if FieldIndex < FItemsListDataSet.FieldCount then
    Result:=FItemsListDataSet.Fields[FieldIndex];
end;

function TrDBCheckListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TrDBCheckListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TrDBRecViewUDB

constructor TrDBRecViewUDB.Create(AOwner: TComponent);
begin
  inherited;
  FRecSQL:=TStringList.Create;
  FRecDataSet:=TQueryIntList.Create(self);
  FRecDataSource:=TDataSource.Create(self);
  FRecAutoLoad:=true;
end;

destructor TrDBRecViewUDB.Destroy;
begin
  FRecDataSource.DataSet:=nil;
  FDataLink.DataSource:=nil;
  FRecDataSet.Close;
  FreeAndNil(FRecDataSource);
  FreeAndNil(FRecDataSet);
  FreeAndNil(FRecSQL);
  inherited;
end;

procedure TrDBRecViewUDB.Loaded;
begin
  inherited;
  FTemplateDataSource:=DataSource;
  if not(csDesigning in ComponentState) then
    if FRecAutoLoad then
      ReloadData;
end;

procedure TrDBRecViewUDB.SetRecSQL(Value: TStrings);
begin
  FRecSQL.Assign(Value);
end;

procedure TrDBRecViewUDB.SetRecConnection(Value: TConnectionUDB);
begin
  FRecConnection:=Value;
  if Assigned(FRecConnection) then
    FRecConnection.FreeNotification(self);
end;

procedure TrDBRecViewUDB.SetRecSQLTryNow(Value: boolean);
begin
  // reset RecSQLTryNow during loading
  if (csLoading in ComponentState) then
    Value:=false;
  FRecSQLTryNow:=Value;
  if FRecSQLTryNow then
    ReloadData
  else
    FRecDataSet.Close;
end;

procedure TrDBRecViewUDB.SetRecDataSetEvent(Value: TDataSetEvent);
begin
  FRecDataSetEvent:=Value;
  if Assigned(FRecDataSet) then
    FRecDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBRecViewUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FRecConnection) and (Operation = opRemove) then
    FRecConnection:=nil;
end;

procedure TrDBRecViewUDB.ReloadData;
var
  i: integer;
  F, T: TField;
  L: TStringList;
begin
  if FRecSQL.Count = 0 then
    Exit;
  try
    FRecDataSet.Close;
    FRecDataSet.SetReadOnly(FRecDataSetReadOnly);
    FRecDataSet.Connection:=FRecConnection;
    if Assigned(FOnGetRecSQL) then
      FOnGetRecSQL(self, RecSQL);
    FRecDataSet.SQL.Assign(FRecSQL);
    FRecDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // keep open in design time to allow set properties in DT dialog
    if Assigned(FTemplateDataSource) and Assigned(FTemplateDataSource.DataSet) then
    begin
      L:=TStringList.Create;
      try
        // copy fields props
        for i:=0 to FTemplateDataSource.DataSet.FieldCount - 1 do
        begin
          F:=FTemplateDataSource.DataSet.Fields[i];
          L.Add(F.FieldName + '=' + F.DisplayLabel);
          T:=FRecDataSet.FindField(F.FieldName);
          if Assigned(T) then
          begin
            T.Visible:=F.Visible;
            T.DisplayWidth:=F.DisplayWidth;
            if T is TNumericField then
              TNumericField(T).DisplayFormat:=TNumericField(F).DisplayFormat;
            if T is TBooleanField then
              TBooleanField(T).DisplayValues:=TBooleanField(F).DisplayValues;
          end;
        end;
        Fields.Assign(L);
      finally
        FreeAndNil(L);
      end;
    end
    else
    begin
      FRecDataSet.SetFieldDefaultFormat;
      FRecDataSet.LoadFieldsProps;
    end;
    FRecDataSource.DataSet:=FRecDataSet;
    FDataLink.DataSource:=FRecDataSource;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TrDBGrid_MSUDB

constructor TrDBGrid_MSUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSource:=TDataSource.Create(self);
  FListAutoLoad:=true;
end;

destructor TrDBGrid_MSUDB.Destroy;
begin
  FListDataSource.DataSet:=nil;
  FListDataSet.Close;
  FreeAndNil(FListDataSource);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListSQL);
  inherited;
end;

procedure TrDBGrid_MSUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadData;
end;

procedure TrDBGrid_MSUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TrDBGrid_MSUDB.SetListConnection(Value: TConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TrDBGrid_MSUDB.SetListSQLTryNow(Value: boolean);
begin
  // reset ListSQLTryNow during loading
  if (csLoading in ComponentState) then
    Value:=false;
  FListSQLTryNow:=Value;
  if FListSQLTryNow then
    ReloadData
  else
    FListDataSet.Close;
end;

procedure TrDBGrid_MSUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TrDBGrid_MSUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TrDBGrid_MSUDB.ReloadData;
begin
  if FListSQL.Count = 0 then
    Exit;
  try
    FListDataSet.Close;
    FListDataSet.SetReadOnly(FListDataSetReadOnly);
    FListDataSet.Connection:=FListConnection;
    if Assigned(FOnGetListSQL) then
      FOnGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // keep open in design time to allow set properties in DT dialog
    FListDataSet.SetFieldDefaultFormat;
    FListDataSet.LoadFieldsProps;
    FListDataSource.DataSet:=FListDataSet;
    DataSource:=FListDataSource;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TrDBGridUDBSorter

function TrDBGridSorter_UDB.SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string): boolean;
var
{$IFDEF USEADO} DS: TCustomADODataSet; {$ENDIF}
{$IFDEF USEDBX} DS: TSimpleDataSet; idxName: string; {$ENDIF}
{$IFDEF USEIB} DS: TIBCustomDataSet; {$ENDIF}
{$IFDEF USEFIB} DS: TpFIBDataSet {$ENDIF}
{$IFDEF USEZeos} DS: TZAbstractDataSet; {$ENDIF}
{$IFDEF USEFD} DS: TFDDataSet; idxName: string; {$ENDIF}
begin
{$IFDEF USEADO}
  if (DataSet is TCustomADODataSet) then
  begin
    DS:=DataSet as TCustomADODataSet;
    DS.Sort:=SortString;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF USEZeos}
  if (DataSet is TZAbstractDataSet) then
  begin
    DS:=DataSet as TZAbstractDataSet;
    DS.SortedFields:=SortString;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF USEDBX}
  if (DataSet is TSimpleDataSet) then
  begin
    DS:=DataSet as TSimpleDataSet;
    // indexname can be max 31 chars long - use hash from fields names
    idxName:=GetIndexNameHash(SortFields + '_' + DescendingFields);
    if DS.IndexDefs.IndexOf(idxName) < 0 then
      DS.AddIndex(idxName, SortFields, [], DescendingFields);
    DS.IndexDefs.Update;
    DS.IndexName:=idxName;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF USEIB}
    Result:= inherited;
{$ENDIF}
{$IFDEF USEFIB}
    Result:= inherited;
{$ENDIF}
{$IFDEF USEFD}
  if (DataSet is TFDDataSet) then
  begin
    DS:=DataSet as TFDDataSet;
    // indexname can be max 31 chars long - use hash from fields names
    idxName:=SortFields + '_' + DescendingFields;
    if DS.IndexDefs.IndexOf(idxName) < 0 then
      DS.AddIndex(idxName, SortFields, '', [], DescendingFields);
    DS.IndexDefs.Update;
    DS.IndexName:=idxName;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
end;

procedure TrDBGridSorter_UDB.DatasetRequery(DataSet: TDataSet);
begin
{$IFDEF USEADO}
  if (DataSet is TCustomADODataSet) then
    TCustomADODataSet(DataSet).Requery()
  else
{$ENDIF}
    inherited;
end;

end.
